/* phase vocoder */
/* Adapted by Richard Dobson to write 1STF SDIF files.
 * NB This is not a normative implementation of 1STF, the definition of which is still
 * under development.
 * 
 * Initial version 0.1 July 2 2000
 */

/* -----------------------------------------------------------------------------
	HISTORY.
	This code is based on pvoc.c from the CARL distribution, as adapted for use in the
	CDP system.
	There is one inherited problem: 
	because the analysis operation puts a rounded number of analysis
	samples into the analysis file, the EXACT original file length
	cannot be restored in the newly synthesised file. While this
	may be annoying, and could no doubt be corrected, it is not
	likely to be of critical importance, since the difference is
	less than one analysis window in samples.

	The fft was 'translated' from fortran to C by Trevor Wishart and Andrew Bentley.
	The fft was made more efficient by Keith Henderson.
	
	the code also incorporates CDP contributions from Martin Atkins and Nick Laviers
	

------------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "pvoc.h"	

#include "pvxutils.h"
#include "riffwav.h"
#include <sdif.h>
#include <sdif-mem.h>





#define	PROG	"PVOCEX"

#define	LOSR	(44100)		
#undef TWOPI			/* Goodness knows why ??? */

#ifndef PI
#define PI	(3.14159265358979323846)
#endif

#define	TWOPI	(2*PI)
/*RWD 3:2000 this rapid report slows things down a lot!: was 0.01 */
#define TIME_INTERVAL	0.5	/*Time between progress report messages*/
#define MONO	(1)		/*In absence of defn. in sfsys.h*/


	Nchans;			/* no of chans */

#ifndef NOOVERCHK
int num_overflows = 0;
double maxsample = 0.0;
double minsample = 0.0;
#endif

void help(void);



void 
logo(void)
{
    printf("\t    ******************************************************\n");
    printf("\t    *                     GEN1STF                         *\n");    
    printf("\t    *                                                    *\n");
    printf("\t    *      A version of Mark Dolson's Phase Vocoder      *\n");
    printf("\t    *            to write SDIF 1STF files                *\n");
    printf("\t    *        Preliminary Version 0.1, RWD June 2000       *\n");
    printf("\t    *                                                    *\n");
    printf("\t    ******************************************************\n\n");
}

void
warpse(float *anal,float *env,int N2,double warp)
{/* spectral envelope detection: this is a very crude peak picking algorithm
	which is used to detect and pre-warp the spectral envelope so that
	pitch transposition can be performed without altering timbre.
	The basic idea is to disallow large negative slopes between
	successive values of magnitude vs. frequency. */

	float	lastmag, mag, nextmag, slope, pkOld,
		eps = (float)(-32.0/N2);
	int	pkcnt, i, j;

	lastmag = *anal;
	mag = *(anal + 2);
	pkOld = lastmag;
	*env = pkOld;
	pkcnt = 1;

	for (i = 1; i <= N2; i++){	/* step thru spectrum*/

		nextmag = (float)( i<N2 ? anal[2*i+2] : 0.0 );

		slope = (float)(pkOld != 0.0 ? (mag-pkOld)/(pkOld*pkcnt) : -10.0 );

						/* look for peaks */
		if ((mag>=lastmag)&&(mag>nextmag)&&(slope>eps)){
			*(env + i) = mag;
			pkcnt--;
			for (j = 1; j <= pkcnt; j++){
				*(env + i - pkcnt + j - 1)
				= (float)(pkOld * (1. + slope * j));
			}
			pkOld = mag;
			pkcnt = 1;
		}
		else pkcnt++;		/* not a peak */

		lastmag = mag;
		mag = nextmag;
	} /* end of scan through spectrum */

	if (pkcnt > 1) {		/* get final peak */
		mag = *(anal + N2*2);
		slope = ((float) (mag - pkOld) / pkcnt);
		*(env + N2) = mag;
		pkcnt--;
		for (j = 1; j <= pkcnt; j++){
			*(env + N2 - pkcnt + j - 1) = pkOld + slope * j;
		}
	}

	for (i = 0; i <= N2; i++){	/* warp spectral env.*/
		j = (int)((float) i * warp);
		if ((j <= N2) && (*(env + i) != 0.))
		*(anal + 2*i) *= *(env + j) / *(env + i);
		else *(anal + 2*i) = 0.0f;
	}
}

void
usage(void)
{
		logo();
		fprintf(stderr,"USAGE: gen1stf [flags] infile outfile\n\n");
		fprintf(stderr,"\ttype gen1stf -h for more information on flags.\n");
		exit(1);
}

void help(void)
{
		fprintf(stderr,"R = input sample rate (automatically read from stdin)\n");
		fprintf(stderr,"F = fundamental frequency (R/256) DONT'T USE -F AND -N\n");
		fprintf(stderr,"N = sample window length for fft (256 unless -F is specified)\n");
		fprintf(stderr,"W = filter overlap factor: {0,1,(2),3} DON'T USE -W AND -M\n");
		fprintf(stderr,"M = analysis window length (N-1 unless -W is specified)\n");
		fprintf(stderr,"D = decimation factor (min((M/(8*T)),(M/8))\n");
#ifdef NOTDEF
		/* none of these needed for analysis only version */
		fprintf(stderr,"L = synthesis window length (M) \n");
		
		fprintf(stderr,"I = interpolation factor (=T*D) \n");
		fprintf(stderr,"T = time-scale factor (1.)\n");
		fprintf(stderr,"P = pitch-scale factor (1.) DON'T USE -T AND -P\n");
		fprintf(stderr,"C = resynthesize odd (1) or even (2) channels only\n");
		fprintf(stderr,"i = resynthesize bandpass filters i thru j only\n");
		fprintf(stderr,"j = resynthesize bandpass filters i thru j only\n");
#endif
		fprintf(stderr,"b = starting sample (0)\n");
		fprintf(stderr,"e = final sample (end of input)\n");
#ifdef NOTDEF
		fprintf(stderr,"w = warp factor for spectral envelope (1.)\n");
		/* RWD for gen1STF: can only do analysis so far! */
		fprintf(stderr,"A:  analysis only: output analysis data (.pvx format)\n");
		/*RWD these options not supported yet */
		fprintf(stderr,"E:  analysis only: output spectral envelope\n");
		fprintf(stderr,"X:  analysis only: output magnitude values\n");
		fprintf(stderr,"S:  synthesis only: input analysis data (.pvx format)\n");
#endif

		fprintf(stderr,"K:  use Kaiser filter instead of hamming\n");
#ifdef NOTDEF
		/*RWD hidden for now */
		fprintf(stderr,"V [filename]:  verbose (summarize on pvoc.s or file)\n");
		fprintf(stderr,"if filename is specified, it must follow all flags\n");
#endif
		exit(1);
}



FUNCTION *gettsf(char *filename)
{	/* Gets pairs of values from the named file and sets up
	   pointer to function header */

	int i=0,len,j,k,l,toggle = 1;
	double sample, *f, *x, *y;
	FUNCTION *p;
	FILE *fp;
	
	/* Open named file for time stretching data input */
	if( (fp = fopen(filename,"r")) == NULL){
		fprintf(stderr,"gen1stf: Cannot open data file\n");
		return (NULL);
	}
	
	/* Read stream of values into buffer */
	while(fscanf(fp,"%lf",&sample) > 0 ){
		if(i == 0){
			f = (double *) malloc(sizeof(double)*BUFSIZ);
			if(f == NULL){
				fprintf(stderr,
					"gen1stf: cannot get memory for buffer\n");
				return (NULL);
			}
			len = BUFSIZ;
		}
		if(i < len-1)
			f[i++]= sample;
		else{
			len += BUFSIZ;
			f = (double *) realloc(f,len*sizeof(double));
			if(f == NULL){
				fprintf(stderr,
					"gen1stf: cannot get memory for buffer\n");
				return (NULL);
			}
			f[i++] = sample;
		}
	}

	/* Rearrange values into arrays */
	f = (double *) realloc(f, i*sizeof(double));
	if(f == NULL){
		fprintf(stderr,
			"gen1stf: cannot get memory for buffer\n");
		return (NULL);
	}
	p = (FUNCTION *) calloc(sizeof(FUNCTION),1);
	if(p == NULL){
		fprintf(stderr,
			"gen1stf: cannot get memory for buffer\n");
		return (NULL);
	}
	if((i%2) != 0){
		fprintf(stderr,"gen1stf: time stretching data not in pairs\n");
		return(NULL);
	}
	x = (double *) malloc(sizeof(double)*i/2);
	y = (double *) malloc(sizeof(double)*i/2);
	for(j=k=l=0; j<i; j++,toggle = 1-toggle){
		if(toggle)
			x[k++] = f[j];
		else
			y[l++] = f[j];
	}
	p->fxval = x;
	p->fyval = y;
	p->flen = i/2;
	p->ftype = "XY pairs";
	p->fname = filename;

	free(f);
	return (p);	
}

void
hamming(float *win,int winLen,int even)
{
	float Pi,ftmp;
	int i;

/***********************************************************
					Pi = (float)((double)4.*atan((double)1.));
***********************************************************/
	Pi = (float)PI;
	ftmp = Pi/winLen;

	if (even) {
		for (i=0; i<winLen; i++)
		*(win+i) = (float)((double).54 + (double).46*cos((double)(ftmp*((float)i+.5))));
		*(win+winLen) = 0.0f;}
	else{	*(win) = 1.0f;
		for (i=1; i<=winLen; i++)
		*(win+i) =(float)((double).54 + (double).46*cos((double)(ftmp*(float)i)));
	}
	return;
}

void
vonhann(float *win,int winLen,int even)
{
	float Pi,ftmp;
	int i;

/***********************************************************
					Pi = (4.*atan(1.));
***********************************************************/
	Pi = (float)PI;
	ftmp = Pi/winLen;

	if (even) {
		for (i=0; i<winLen; i++)
		*(win+i) = (float)(.5 + .5 *cos(ftmp*((double)i+.5)));
		*(win+winLen) = 0.0f;
	}
	else{	*(win) = 1.0f;
		for (i=1; i<=winLen; i++)
		*(win+i) =(float)(.5 + .5 *cos(ftmp*(double)i));
	}
	return;
}

/******************** KAISER WINDOW ***************/
double besseli( double x)
{
    double ax, ans;
    double y;

    if (( ax = fabs( x)) < 3.75)     {
	y = x / 3.75;
	y *= y;
	ans = (1.0 + y * ( 3.5156229 +
			  y * ( 3.0899424 +
				y * ( 1.2067492 +
				      y * ( 0.2659732 +
					    y * ( 0.360768e-1 +
						  y * 0.45813e-2))))));
    }
    else {
	y = 3.75 / ax;
	ans = ((exp ( ax) / sqrt(ax))
	    * (0.39894228 +
	       y * (0.1328592e-1 +
		    y * (0.225319e-2 +
			 y * (-0.157565e-2 +
			      y * (0.916281e-2 +
				   y * (-0.2057706e-1 +
					y * (0.2635537e-1 +
					     y * (-0.1647633e-1 +
						  y * 0.392377e-2)))))))));
    }
    return ans;
}

//courtesy of Csound....

void kaiser(float *win,int len,double beta)
{
	float *ft = win;
	double i,xarg = 1.0;	//xarg = amp scalefactor
	for (i = -len/2.0 + .1 ; i < len/2.0 ; i++)
		*ft++ = (float) (xarg *
              besseli((beta * sqrt(1.0-pow((2.0*i/(len - 1)), 2.0)))) /
              besseli( beta));
   // assymetrical hack: sort out first value!
   win[0] = win[len-1];
}

float
*float_array(int nnn)
{	/* set up a floating point array length nnn. */
	float *ptr;
	ptr = (float *) calloc(nnn,sizeof(float));
	if(ptr==NULL){
		fprintf(stderr,"gen1stf: insufficient memory\n");
		exit(1);
	}
	return (ptr);
}

void
malerr(char *str, int ex)
{
	fprintf(stderr, "%s\n", str);
	exit(1);
}




void
main(int argc, char *argv[])
{
	double	rratio;
	float	*input,		/* pointer to start of input buffer */
		*output,	/* pointer to start of output buffer */
		*anal,		/* pointer to start of analysis buffer */
		*syn,		/* pointer to start of synthesis buffer */
		*banal,		/* pointer to anal[1] (for FFT calls) */
		*bsyn,		/* pointer to syn[1]  (for FFT calls) */
		*nextIn,	/* pointer to next empty word in input */
		*nextOut,	/* pointer to next empty word in output */
		*analWindow,	/* pointer to center of analysis window */
		*synWindow,	/* pointer to center of synthesis window */
		*maxAmp,	/* pointer to start of max amp buffer */
		*avgAmp,	/* pointer to start of avg amp buffer */
		*avgFrq,	/* pointer to start of avg frq buffer */
		*env,		/* pointer to start of spectral envelope */
		*i0,		/* pointer to amplitude channels */
		*i1,		/* pointer to frequency channels */
		*oi,		/* pointer to old phase channels */
		*oldInPhase,	/* pointer to start of input phase buffer */
		*oldOutPhase;	/* pointer to start of output phase buffer */

	int	m, n;

	int	N = 0,		/* number of phase vocoder channels (bands) */
		M = 0,		/* length of analWindow impulse response */
		L = 0,		/* length of synWindow impulse response */
		D = 0,		/* decimation factor (default will be M/8) */
		I = 0,		/* interpolation factor (default will be I=D)*/
		W = -1, 	/* filter overlap factor (determines M, L) */
		F = 0,		/* fundamental frequency (determines N) */
		F2 = 0,		/* F/2 */
		analWinLen,	/* half-length of analysis window */
		synWinLen;	/* half-length of synthesis window */

	long	tmprate,	/* temporary variable */
		sampsize,	/* sample size for output file */
		origrate,	/* sample rate of file analysed */
		outCount,	/* number of samples written to output */
		ibuflen,	/* length of input buffer */
		obuflen,	/* length of output buffer */
		nI = 0,		/* current input (analysis) sample */
		nO,		/* current output (synthesis) sample */
		nMaxOut,	/* last output (synthesis) sample */
		nMin = 0,	/* first input (analysis) sample */
		nMax = 100000000;	/* last input sample (unless EOF) */
/***************************** 6:2:91  OLD CODE **************
						long	origsize;
*******************************NEW CODE **********************/
	//long	origsize = 0;	/* sample type of file analysed */

	char	ch;		/* needed for crack (commandline interpreter)*/

	FILE	*fp;		/* auxiliary output file (-V option) */
	
	RIFFWAV_STRUCT *insndfile = NULL,*outsndfile = NULL;
	FILE *sdiffile = NULL;
	SDIFresult r;
	SDIF_FrameHeader head;
	SDIF_MatrixHeader SDIF_MatrixHeader;
	float framedur,framepos,frameincr,sdif_nyquist;
	long framesize;




	int flags;
	float *analWindowBase;
	float *synWindowBase;


	float	beta = 6.8f,	/* parameter for Kaiser window */
		real,		/* real part of analysis data */
		imag,		/* imaginary part of analysis data */
		mag,		/* magnitude of analysis data */
		phase,		/* phase of analysis data */
		angleDif,	/* angle difference */
		RoverTwoPi,	/* R/D divided by 2*Pi */
		TwoPioverR,	/* 2*Pi divided by R/I */
		sum,		/* scale factor for renormalizing windows */
		ftot = 0.0f,	/* scale factor for calculating statistics */
		rIn,		/* decimated sampling rate */
		rOut,		/* pre-interpolated sampling rate */
		invR,		/* 1. / srate */
		time,		/* nI / srate */
		tvx0,		/* current x value of time-var function */
		tvx1,		/* next x value of time-var function */
		tvdx,		/* tvx1 - tvx0 */
		tvy0,		/* current y value of time-var function */
		tvy1,		/* next y value of time-var function */
		tvdy,		/* tvy1 - tvy0 */
		frac,		/* tvdy / tvdx */
		warp = 0.0f,	/* spectral envelope warp factor */
		R = 0.0f,		/* input sampling rate */
		P = 1.0f,		/* pitch scale factor */
		Pinv,		/* 1. / P */
		T = 1.0f;		/* time scale factor ( >1 to expand)*/

	int	i,j,k,		/* index variables */
		Dd,		/* number of new inputs to read (Dd <= D) */
		Ii,		/* number of new outputs to write (Ii <= I) */
		N2,		/* N/2 */
		NO,		/* synthesis NO = N / P */
		NO2,		/* NO/2 */
		IO,		/* synthesis IO = I / P */
		IOi,		/* synthesis IOi = Ii / P */
		Mlen,
		Mf = 0,		/* flag for even M */
		Lf = 0,		/* flag for even L */
		Dfac,
		flag = 0,		/* end-of-input flag */
		C = 0,		/* flag for resynthesizing even or odd chans */
		Ci = 0,		/* flag for resynthesizing chans i to j only */
		Cj = 0,		/* flag for resynthesizing chans i to j only */
		CC = 0,		/* flag for selected channel resynthesis */
		V = 0,		/* verbose (summarize analysis) output flag */
		K = 0,		/* flag for Kaiser window */
		/*RWD for gen1STF always do analysis */
		A = 1,		/* flag for analysis only */
		X = 0,		/* flag for magnitude output */
		E = 0,		/* flag for spectral envelope output */
		S = 0,		/* flag for synthesis only */
		tvflg = 0,	/* flag for time-varying time-scaling */
		tvnxt,		/* counter for stepping thru time-var func */
		tvlen,		/* length of time-varying function */
		rv;		/* return value from fgetfloat */
	float	srate,		/* sample rate from header on stdin */
		arate;		/* sample rate for header on stdout if -A */
	float	timecheck = 0.0f;

	

	if(SDIF_Init() != ESDIF_SUCCESS){
		fprintf(stderr,"\nunable to startup SDIF");
		exit(1);
	}


	if(argc==2 && !strcmp(argv[1],"-h"))
		help();
	m = 0;
	for(n=1;n<argc;n++) {	  
		if(m && *argv[n] !='-'){   /*don't count filenames starting with '-' !  */
			/*fprintf(stderr,"pvoc WARNING: filename starts with '-' ! \n"); */
			break;
		}
		if(*argv[n] =='-'){
			m++;
		}
	}
	if(argc-m < 3)					 /*RWD I don't trust this */
		usage();

/* call crack to interpret commandline */
	flags = 0;
	/*RWD for gen1STF exclude A option - always done */
    while ((ch = crack(argc,argv,"N|M|L|D|I|R|P|T|b|e|i|j|w|W|F|C|XESKVh",0)) != 0){
		
		switch (ch) {
			case 'N': N    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'M': M    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'L': L    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'D': D    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'I': I    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'R': R    = (float) sfexpr(arg_option,1.0); break;
			case 'P': P    = (float) sfexpr(arg_option,1.0); break;
			case 'T': T    = (float) sfexpr(arg_option,1.0); break;
			case 'W': W    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'F': F    = (int) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'C': C    = (int) (sfexpr(arg_option,1.0) + 0.5);CC=1; break;
			case 'i': Ci   = (int) (sfexpr(arg_option,1.0)+ 0.5); CC=1; break;
			case 'j': Cj   = (int) (sfexpr(arg_option,1.0)+ 0.5); CC=1; break;
			case 'b': nMin = (long) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'e': nMax = (long) (sfexpr(arg_option,1.0) + 0.5); break;
			case 'w': 
				if (arg_option[0] == 0) 
					warp = -1.0f;
				else 
					warp = (float) sfexpr(arg_option,1.0);
				break;
#ifdef NOTDEF
			case 'A': A = 1; break;
#endif
			case 'E':
				/*RWD*/
				fprintf(stderr,"\nSorry: -E option not yet supported for gen1stf");
				exit(1);

				/*A = 1; E = 1; break;*/
			case 'X': 
				/*RWD*/
				fprintf(stderr,"\nSorry: -X option not yet supported for gen1stf");
				exit(1);
				/*A = 1; X = 1; break;*/
#ifdef NOTDEF
			case 'S': S = 1; break;
#endif
			case 'K':
				K = 1; 
				/*fprintf(stderr, "\npvoc: Kaiser window not implemented!\n");*/
				break;
			case 'V': 
				V = 1; 
				break;
			case 'h': usage();	/* then exits normally */
			default: usage();	/* this exits with error */
		}
		flags++;
	}
	argv+= flags;
	argc-= flags;
	if(argc < 2){
		fprintf(stderr,"\ninsufficient arguments");
		usage();
		
	}

		/* infile is soundfile */		
		insndfile = wavNew();
		if(insndfile==NULL){
			puts("\nno memory!");
			exit(1);
		}
		if(wavOpenRead(insndfile, argv[1], 0) != RIFFWAV_E_NOERROR){
			fprintf(stderr,"\nUnable to open input soundfile %s",argv[1]);
			exit(1);
		}



		srate = (float) insndfile->fmt.nSamplesPerSec;	/* get input srate */
		if (R == 0.)
			R = srate;
		srate = R;
		
		Nchans = insndfile->fmt.nChannels;
		/*RWD um, maybe we can't have stereo SDIF files, just lots of mono streams...*/
		if( Nchans != MONO){
			fprintf(stderr,"gen1stf: input soundfile is not mono\n");
			exit(1);
		}




	if ((N != 0) && (F != 0))
		fprintf(stderr,"gen1stf: warning - don't specify both N and F\n");
	if ((N == 0) && (F == 0))
		N = 256;
	else
		if (F != 0)
			N = (int)((float) R / F);
	N = N  + N%2;	/* Make N even */
	N2 = N / 2;
	if (N2 > 16384){
		fprintf(stderr,"gen1stf: N too large\n");
		exit(1);
	}

	F = (int)((float) R / N);
	F2 = F / 2;

	if (W != -1)
	 	if (M != 0)
		fprintf(stderr,"gen1stf: warning - don't specify both M and W\n");
		else
			switch(W){
				case 0:	M = 4*N;
					break;
				case 1:	M = 2*N;
					break;
				case 2: M = N;
					break;
				case 3: M = N2;
					break;
				default:
		fprintf(stderr,"gen1stf: warning - invalid W ignored\n");
					break;
			}
	M = (M != 0 ? M : N);
	Mf = 1 - M%2;

	L = (L != 0 ? L : M);
	Lf = 1 - L%2;

	if (M < 7)
		fprintf(stderr,"gen1stf: warning - M is too small\n");

	ibuflen = 4 * M;
	obuflen = 4 * L;

	if (W == -1)
		W = (int)(3.322 * log10((double)(4. * N) / M));/* cosmetic */

	if ((A == 1) && (S == 1)) {
		fprintf(stderr,"gen1stf: can't specify both -A and -S\n");
		exit(1);
		}

	if (Cj == 0) Cj = N2;
	if (Cj > N2) Cj = N2;
	if (Ci < 0) Ci = 0;


	if(P != 1.0){
		if(T != 1.0)
			fprintf(stderr,"gen1stf: warning - don't specify both T and P\n");

		if(tvflg){
			fprintf(stderr,"gen1stf: can't specify -P with function\n");
			exit(1);
		}

		T = P;	/* pitch change is time change plus resamp */
	}
	if ((D != 0) && (T != 1.))
		fprintf(stderr,"gen1stf: warning - don't specify both D and T\n");

	if (T <= 0.){
		fprintf(stderr,"gen1stf: invalid T = %f\n",T);
		exit(1);
	}


#ifdef OLD
	D = (D != 0 ? D : M/(8.0*max(1.0,T)) );	/* MCA fix - wrong types!! */
#endif
	D = (int)((D != 0 ? D : M/(8.0*(T > 1.0 ? T : 1.0))));

	if (D == 0){
		fprintf(stderr,"gen1stf: warning - T greater than M/8 \n");
		D = 1;
	}

	I = (int)(I != 0 ? I : (float) T*D );

	if (I == 0){
		fprintf(stderr,"gen1stf: T*D (or P*D) < 1 - increase M\n");
		exit(1);
	}

	if (tvflg){
		T = tvy0;	/* original T was maximum - used to set I */
		D = (int)((float) I / T);
		if (D < 1){
		fprintf(stderr,"gen1stf: warning - can't expand by %f\n",T);
			D = 1;
		}
	}

	if (tvflg)
		if (warp != 0.)
			warp = T;	/* warp varies with T */

	T = ((float) I / D);

	if (P != 1.)
		P = T;

	NO = (int)((float) N / P);	/* synthesis transform will be NO points */
	NO = NO + NO%2;		/* make NO even */

	NO2 = NO / 2;

	P = ((float) N / NO);	/* ideally, N / NO = I / D = pitch change */
	Pinv = (float)(1.0/ P);

	if (warp == -1.)
		warp = P;
	if ((E == 1) && (warp == 0.))
		warp = 1.0f;


	if ((P != 1.) && (P != T))
		 fprintf(stderr,"gen1stf: warning P=%f not equal to T=%f\n",P,T);

	IO = (int)((float) I / P);

	nMax -= nMin;


	if (V) {
		fprintf(fp,"\nN: %d  M: %d  L: %d",N,M,L); 
		fprintf(fp,"  D: %d  I: %d  F: %d",D,I,F);
		fprintf(fp,"  R: %7.1f  P: %5.2f  T: %5.2f\n",R,P,T);
		if (CC)
			fprintf(fp,"C: %d    i: %d    j: %d\n",C,Ci,Cj);
		if (K)
			fprintf(fp,"---Kaiser Window---\n");
	}

	/*RWD: now, we can open outfile */
	if(A){
		/* create gen1stf file - mono only, for now */
	
		/* open SDIF outfile */
		if (r = SDIF_OpenWrite(argv[2], &sdiffile)) {
			fprintf (stderr, "Error creating %s: %s\n",argv[2],
			SDIF_GetErrorString(r));
			fclose(sdiffile);
			exit(1);
		}
    
		SDIF_Copy4Bytes(head.frameType,"1STF");
		head.streamID = SDIF_UniqueStreamID();
		head.matrixCount = 2;			/* info matrix and data matrix */
		/* frame size =  sizeof(info matrix) + sizeof (data matrix) + frame header fields*/
		/* INFO matrix */
		/* add main header fields */
		head.size = sizeof(sdif_float64) + 2 * sizeof(sdif_int32);		
		/* add header block of a matrix */
		head.size += 4 * sizeof(sdif_int32);
		/* add info fields */
		head.size += 2 * sizeof(sdif_float32);
		/* no padding needed for info matrix */
		/* now add size of data matrix */
		/* add header block of data matrix */
		head.size += 4 * sizeof(sdif_int32);

		/*RWD speculation: SDIF requre pow-of-two complex spectrum...*/

		head.size += 2 * sizeof(sdif_float32) * (N2 + 1);
		/* no padding bytes */

		framedur = (float)N / srate;
		framepos = framedur * 0.5;		/* times supposed to give centre of window */
		framesize = (N2+1) * 2* sizeof(float);
		sdif_nyquist = srate * 0.5;
		/* frameincr =  overlap / srate */
		frameincr = (float)D / srate;

	}




	printf("\ngen1stf - analysis beginning\n");	

	/* set up analysis window: The window is assumed to be symmetric
		with M total points.  After the initial memory allocation,
		analWindow always points to the midpoint of the window
		(or one half sample to the right, if M is even); analWinLen
		is half the true window length (rounded down). Any low pass
		window will work; a Hamming window is generally fine,
		but a Kaiser is also available.  If the window duration is
		longer than the transform (M > N), then the window is
		multiplied by a sin(x)/x function to meet the condition:
		analWindow[Ni] = 0 for i != 0.  In either case, the
		window is renormalized so that the phase vocoder amplitude
		estimates are properly scaled.  The maximum allowable
		window duration is ibuflen/2. */


	analWindow = float_array(M+Mf);

	analWindowBase = analWindow;

	analWindow += (analWinLen = M/2);

	if (K){	
		/*kaiser_(&M,analWindow,&analWinLen,(int*)1,&beta);*/
		/* RWD this function does whole window, not one half */
		kaiser(analWindowBase,/*analWinLen*/M,beta);			/*RWD */
		
	}

	else	
		hamming(analWindow,analWinLen,Mf);
	if(K==0)  /*RWD see above */
		for (i = 1; i <= analWinLen; i++)
			*(analWindow - i) = *(analWindow + i - Mf);

	if (M > N) {
		if (Mf)
		*analWindow *=(float)( (double)N * sin((double)PI*.5/N) /(double)( PI*.5));
		for (i = 1; i <= analWinLen; i++) 
			*(analWindow + i) *=(float)
			((double)N * sin((double) (PI*(i+.5*Mf)/N) / (PI*(i+.5*Mf))));
		for (i = 1; i <= analWinLen; i++)
			*(analWindow - i) = *(analWindow + i - Mf);
	}

	sum = 0.0f;
	for (i = -analWinLen; i <= analWinLen; i++)
		sum += *(analWindow + i);

	sum = (float)(2.0 / sum);		/*factor of 2 comes in later in trig identity*/

	for (i = -analWinLen; i <= analWinLen; i++)
		*(analWindow + i) *= sum;

	/* set up synthesis window:  For the minimal mean-square-error
		formulation (valid for N >= M), the synthesis window
		is identical to the analysis window (except for a
		scale factor), and both are even in length.  If N < M,
		then an interpolating synthesis window is used. */

	synWindow = float_array(L+Lf);
	synWindowBase = synWindow;
	synWindow += (synWinLen = L/2);

	if (M <= N){
		if (K) 	
			/*kaiser_(&M,synWindow,&synWinLen,(int*)1,&beta);*/
			kaiser(synWindowBase,/*synWinLen*/M,beta);			/*RWD */
		else
			hamming(synWindow,synWinLen,Lf);
		if(K==0)
			for (i = 1; i <= synWinLen; i++)
				*(synWindow - i) = *(synWindow + i - Lf);

		for (i = -synWinLen; i <= synWinLen; i++)
			*(synWindow + i) *= sum;

		sum = 0.0f;
		for (i = -synWinLen; i <= synWinLen; i+=I)
			sum += *(synWindow + i) * *(synWindow + i);

		sum = (float)(1.0/ sum);

		for (i = -synWinLen; i <= synWinLen; i++)
			*(synWindow + i) *= sum;}
	else {	
		if (K) 	  /*RWD added: hope this is correct! */
			/*kaiser_(&M,synWindow,&synWinLen,(int*)1,&beta);*/
			kaiser(synWindowBase,/*synWinLen*/M,beta);			/*RWD */
		else

			hamming(synWindow,synWinLen,Lf);
		if(K==0)
			for (i = 1; i <= synWinLen; i++)
				*(synWindow - i) = *(synWindow + i - Lf);

		if (Lf)
			*synWindow *= (float)((double)IO * sin((double) (PI*.5/IO)) / (double)(PI*.5));
		for (i = 1; i <= synWinLen; i++) 
			*(synWindow + i) *=(float)
			((double)IO * sin((double) (PI*(i+.5*Lf)/IO)) /(double) (PI*(i+.5*Lf)));
		for (i = 1; i <= synWinLen; i++)
			*(synWindow - i) = *(synWindow + i - Lf);

		sum = (float)(1.0/sum);

		for (i = -synWinLen; i <= synWinLen; i++)
			*(synWindow + i) *= sum;
	}
      
	/* set up input buffer:  nextIn always points to the next empty
		word in the input buffer (i.e., the sample following
		sample number (n + analWinLen)).  If the buffer is full,
		then nextIn jumps back to the beginning, and the old
		values are written over. */

	input = float_array(ibuflen);

	nextIn = input;

	/* set up output buffer:  nextOut always points to the next word
		to be shifted out.  The shift is simulated by writing the
		value to the standard output and then setting that word
		of the buffer to zero.  When nextOut reaches the end of
		the buffer, it jumps back to the beginning.  */

	output =	float_array(obuflen);

	nextOut = output;
	/* set up analysis buffer for (N/2 + 1) channels: The input is real,
		so the other channels are redundant. oldInPhase is used
		in the conversion to remember the previous phase when
		calculating phase difference between successive samples. */

	anal =		float_array(N+2);
	banal = anal + 1;

	oldInPhase =	float_array(N2+1);
	maxAmp =	float_array(N2+1);
	avgAmp =	float_array(N2+1);
	avgFrq =	float_array(N2+1);
	env =		float_array(N2+1);
	/* set up synthesis buffer for (N/2 + 1) channels: (This is included
		only for clarity.)  oldOutPhase is used in the re-
		conversion to accumulate angle differences (actually angle
		difference per second). */

	syn =		float_array(NO+2);
	bsyn = syn + 1;

	oldOutPhase =	float_array(NO2+1);

	/* initialization: input time starts negative so that the rightmost
		edge of the analysis filter just catches the first non-zero
		input samples; output time is always T times input time. */

	
	if(R == 0) 
		exit(1);
				/* discard samples at start */
	for (i = 0; i < nMin; i++) 
		/*fgetfloat(nextIn,ifd);*/
		if(wavSeek(insndfile,nMin) != nMin){
			fprintf(stderr,"\nError seeking forward in infile");
			exit(1);
		}

	outCount = 0;
	rIn = ((float) R / D);
	rOut = ((float) R / I);
	invR =((float) 1. / R);
	RoverTwoPi = (float)(rIn / TWOPI);
	TwoPioverR = (float)(TWOPI / rOut);
	nI = -(analWinLen / D) * D;	/* input time (in samples) */
	nO = (long)((float) T/P * nI);	/* output time (in samples) */
	Dd = analWinLen + nI + 1;	/* number of new inputs to read */
	Ii = 0;				/* number of new outputs to write */
	IOi = 0;
	flag = 1;

	/* main loop:  If nMax is not specified it is assumed to be very large
		and then readjusted when fgetfloat detects the end of input. */
	while(nI < (nMax + analWinLen)){
		int got;
		time = nI * invR;
		if(time>timecheck){
			printf("\rInput time: %5.2f secs\t Output time: %5.2f secs",
			 time, nO * invR);
			fflush(stdout);
			timecheck = (float)(timecheck + TIME_INTERVAL);
		}
	
		if (S == 1){
			/* watch this space....*/
		}
		else {		/* prepare for analysis: read next Dd input values */

			static float *sbuf = 0;
			static int sblen = 0;
			int got, tocp;
			float *sp;

			if(sblen < Dd) {
				if(sbuf != 0)
					free(sbuf);
				if((sbuf = (float *)malloc(Dd*sizeof(float))) == 0) {
					fprintf(stderr, "gen1stf: can't allocate float buffer\n");
					exit(1);
				}
				sblen = Dd;
			}
			
			if((got = wavReadFloatFrames(insndfile,sbuf,Dd)) < 0){
				fprintf(stderr,"\nError reading infile");
				exit(1);
			}
			if(got < Dd)
				Dd = got;
			sp = sbuf;
			tocp = min(got, input+ibuflen-nextIn);
			got -= tocp;
			while(tocp-- > 0)
				*nextIn++ = *sp++;

			if(got > 0) {
				nextIn -= ibuflen;
				while(got-- > 0)
					*nextIn++ = *sp++;
			}
			if (nextIn >= (input + ibuflen))
				nextIn -= ibuflen;
		}

		if (nI > 0)
			for (i = Dd; i < D; i++){	/* zero fill at EOF */
				*(nextIn++) = 0.0f;
				if (nextIn >= (input + ibuflen))
					nextIn -= ibuflen;
			}
	/* analysis: The analysis subroutine computes the complex output at
		time n of (N/2 + 1) of the phase vocoder channels.  It operates
		on input samples (n - analWinLen) thru (n + analWinLen) and
		expects to find these in input[(n +- analWinLen) mod ibuflen].
		It expects analWindow to point to the center of a
		symmetric window of length (2 * analWinLen +1).  It is the
		responsibility of the main program to ensure that these values
		are correct!  The results are returned in anal as succesive
		pairs of real and imaginary values for the lowest (N/2 + 1)
		channels.   The subroutines fft and reals together implement
		one efficient FFT call for a real input sequence.  */


		for (i = 0; i < N+2; i++) *(anal + i) = 0.0f;	/*initialize*/

		j = (nI - analWinLen - 1 + ibuflen) % ibuflen;	/*input pntr*/

		k = nI - analWinLen - 1;			/*time shift*/
		while (k < 0)
			k += N;
		k = k % N;
		for (i = -analWinLen; i <= analWinLen; i++) {
			if (++j >= ibuflen)
				j -= ibuflen;
			if (++k >= N)
				k -= N;
			*(anal + k) += *(analWindow + i) * *(input + j);
		}

		/* RWD BIG QUESTION! What format does SDIF expect?
		 *  in sdif2mp4, it requires a full complex spectrum, with re[Nyquist] at size/2
		 */

		fft_(anal,banal,1,N2,1,-2);
		reals_(anal,banal,N2,-2);

	
		/* write out info matrix, then data matrix of real:imag.
		 *  It ~appears~ that the 1STF frame should contain a 'half-complex'
		 * array of real-imaginary, from DC up to Nyquist, ~inclusive~.
		 * it is ~expected~ that there will be (N/2)+1 rows.
		 * here the first bins contain the values for the fundamental freq of the analysis,
		 * and therefore we optimistically add zero-value fields for the DC bin.
		 * until we have a normative 1STF synthesizer, it is difficult to be sure
		 * if this is correct! This setup currently gets the best results using sdif.saol
		 * (but still far from correct).
		 */
		if (A == 1){
			long zero = 0;
			head.time = framepos;
			SDIF_WriteFrameHeader(&head,sdiffile);
			/* setup info matrix */
			SDIF_MatrixHeader.rowCount = 1;
			SDIF_MatrixHeader.columnCount = 2;
			SDIF_Copy4Bytes(SDIF_MatrixHeader.matrixType,"ISTF");
			SDIF_MatrixHeader.matrixDataType = SDIF_FLOAT32;			
			SDIF_WriteMatrixHeader(&SDIF_MatrixHeader,sdiffile);
			SDIF_Write4(&sdif_nyquist,1,sdiffile);
			SDIF_Write4(&framedur,1,sdiffile);
			/*setup data matrix */
			SDIF_MatrixHeader.rowCount = N2 + 1 ;
			SDIF_MatrixHeader.columnCount = 2;
			SDIF_Copy4Bytes(SDIF_MatrixHeader.matrixType,"1STF");
			SDIF_MatrixHeader.matrixDataType = SDIF_FLOAT32;			
			SDIF_WriteMatrixHeader(&SDIF_MatrixHeader,sdiffile);
			/* write DC components ?*/
			SDIF_Write4(&zero,1,sdiffile);
			SDIF_Write4(&zero,1,sdiffile);
			/*write fund up to Nyquist*/
			for(i=0;i < N;i+=2){
				SDIF_Write4(&(anal[i]),1,sdiffile);
				SDIF_Write4(&(anal[i+1]),1,sdiffile);
			}
		}
				
		if (flag)	/* flag means do this operation only once */
			if ((nI > 0) && (Dd < D)){	/* EOF detected */
				flag = 0;
				nMax = nI + analWinLen - (D - Dd);
			}
		

		/*	D = some_function(nI);		for variable time-scaling */
		/*	rIn = ((float) R / D);		for variable time-scaling */
		/*	RoverTwoPi =  rIn / TwoPi;	for variable time-scaling */

		nI += D;				/* increment time */
		nO += IO;

		/* Dd = D except when the end of the sample stream intervenes */

		Dd = min(D, max(0, D+nMax-nI-analWinLen));


		if (nO > (synWinLen + I))
			Ii = I;
		else
			if (nO > synWinLen)
				Ii = nO - synWinLen;
			else {
				Ii = 0;
				for (i=nO+synWinLen; i<obuflen; i++)
					if (i > 0)
						*(output+i) = 0.0f;
			}
		IOi = (int)((float) Ii / P);

		/* step SDIF frametime to centre of next window */
		framepos += frameincr;

	}	/* End of main while loop */

	if (V) {
		if (ftot != 0.)
			ftot = (float)(1.0/ ftot);

		fprintf(fp,"\n   i         band       max amp     ");
		fprintf(fp,"avg amp    avg frq\n\n");
		for (i = 0; i <= N/2; i++) fprintf(fp,
			"%4d   %5d - %5d   %8.5f    %8.5f   %8.1f\n",
			2*i+1, i*F-F2, i*F+F2, *(maxAmp+i), *(avgAmp+i)*ftot, 
			*(avgFrq+i)*ftot);
		fprintf(fp,"\n");
	}
#ifdef NOTDEF
	if (A != 1) {
		nMaxOut = (long)((float) T / P * nMax);
		while (outCount <= nMaxOut){
			int todo = min(nMaxOut-outCount, output+obuflen-nextOut);
			if(todo == 0)
				break;			
			if(wavWriteFloatFrames(outsndfile,nextOut,todo) < todo){
				fprintf(stderr,"\nWarning: error writing data to outfile");
				/* in case we can recover something anyway*/
				goto epilog;
			}
			outCount += todo;
			nextOut += todo;
			if (nextOut >= (output + obuflen))
				nextOut -= obuflen;
		}
	}
#endif
epilog:
	if(insndfile) {
		wavClose(insndfile);
		wavFree(insndfile);
	}
	if(outsndfile){
		wavClose(outsndfile);
		wavFree(outsndfile);
	}
	if(sdiffile)
		fclose(sdiffile);
#ifdef NOTDEF
	if(p_pvdata)
		free(p_pvdata);
	if(p_wfx)
		free(p_wfx);
#endif

#ifndef NOOVERCHK
	if(num_overflows > 0) {
		fprintf(stderr, "Warning: %d samples overflowed, and were clipped\n", num_overflows);
		fprintf(stderr, "\tmaximum sample was %f\n", maxsample);
		fprintf(stderr, "\tminimum sample was %f\n", minsample);
		if(-minsample > maxsample)
			maxsample = -minsample;
/*RWD: avoid div by zero...*/
		if(maxsample!=0.)
			fprintf(stderr, "You should reduce source level to avoid clipping: use gain of <=%lf\n", 1.0/maxsample);
	}
#endif
	exit(0);
}

